#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License
import os
import signal
import subprocess
import sys
import threading
import time
from datetime import datetime
from functools import wraps
from pathlib import Path
from typing import List, Optional

import psutil

from sparglim.config.configer import SparkEnvConfiger
from sparglim.exceptions import DaemonError, UnconfigurableError
from sparglim.log import logger
from sparglim.server.tailer import Tailer
from sparglim.utils import get_scala_version, get_spark_version, port_is_used


class Daemon:
    # This daemon is designed for spark connect server in the frontend
    # Run the connect server in a container and monitor its state, if it is down, restart it
    # Stop the daemon when the KeyboardInterrupt is raised
    # This will not start duplicate connect server, unless using different root_dir and port
    ENV_MASTER_MODE = "SPARGLIM_SERVER_MODE"
    DEFAULT_MODE = "local"
    SPARK_IDENT_STRING = "sparglim-connect-server"  # SPARK_IDENT_STRING

    def __init__(
        self,
        mode: Optional[str] = None,
        root_dir: str = "./",
        *,
        k8s_config_path: Optional[str] = None,
    ):
        self.mode = mode or os.getenv(self.ENV_MASTER_MODE, self.DEFAULT_MODE)

        self.root_dir = Path(root_dir)

        self.log_dir = self.root_dir / "logs"  # SPARK_LOG_DIR
        self.pid_dir = self.root_dir / "pids"  # SPARK_PID_DIR
        self.log_glob_format = f"spark-{self.SPARK_IDENT_STRING}-*.out"
        self.pid_glob_format = f"spark-{self.SPARK_IDENT_STRING}-*.pid"
        self.daemon_pid_file = self.pid_dir / f"daemon.pid"

        self.spark_cmd_dir = Path(f"{os.getenv('SPARK_HOME')}/sbin")
        self.start_cmd_path = self.spark_cmd_dir / "start-connect-server.sh"
        if not self.start_cmd_path.exists():
            raise UnconfigurableError(
                f"{self.start_cmd_path.name} not found in $SPARK_HOME/sbin, check your SPARK_HOME"
            )

        SPARK_VERSION = os.getenv("SPARK_VERSION", get_spark_version())
        SCALA_VERSION = os.getenv("SCALA_VERSION", get_scala_version())
        self.daemon_package = []
        if SPARK_VERSION and SCALA_VERSION:
            # Specify the spark-connect package or use the default one
            self.daemon_package.append(
                f"org.apache.spark:spark-connect_{SCALA_VERSION}:{SPARK_VERSION}"
            )
        self.builder = SparkEnvConfiger().config_connect_server(
            self.mode, k8s_config_path=k8s_config_path
        )

        self._tailer: Optional[Tailer] = None
        self._daemon: Optional[threading.Thread] = None
        self.is_stopping: bool = False

    @property
    def config(self) -> List[str]:
        return [f"--conf {k}={v}" for k, v in self.builder.get_all().items()]

    @property
    def packages(self) -> List[str]:
        return [f"--packages {p}" for p in self.daemon_package]

    @property
    def pid_file(self) -> Optional[Path]:
        for pid_file in self.pid_dir.glob(self.pid_glob_format):
            return pid_file
        return None

    @property
    def log_file(self) -> Optional[Path]:
        for log_file in self.log_dir.glob(self.log_glob_format):
            return log_file
        return None

    @property
    def pid(self) -> Optional[int]:
        if not self.pid_file:
            return None
        return int(self.pid_file.read_text())  # type: ignore

    @property
    def port(self) -> int:
        return int(self.builder.get_all().get("spark.connect.grpc.binding.port", "15002"))

    def start_and_daemon(self):
        # Start spark connect server and daemon it
        # If it is already running, just daemon it
        if not self.pid or not psutil.pid_exists(self.pid):
            self._start()
        if self.daemon_pid_file.exists():
            d_pid = self.daemon_pid_file.read_text()
            if psutil.pid_exists(int(d_pid)):
                logger.warning(f"Daemon is already running, pid: {d_pid}, exit")
                return
        else:
            self.daemon_pid_file.write_text(str(os.getpid()))
        self._launch_daemon()
        self._wait_exit()

    def _start(self):
        self._rotate_log()
        logger.info(f"Start connect server, work dir: {self.root_dir.absolute()}")
        if port_is_used(self.port):
            logger.error(f"Port {self.port} is already in use")
            raise UnconfigurableError(f"Port {self.port} is already in use")
        env = os.environ.copy()
        env["SPARK_LOG_DIR"] = self.log_dir.absolute().as_posix()
        env["SPARK_PID_DIR"] = self.pid_dir.absolute().as_posix()
        env["SPARK_IDENT_STRING"] = self.SPARK_IDENT_STRING
        # We rotate log by ourselves
        env["SPARK_LOG_MAX_FILES"] = "1"
        cmd = [self.start_cmd_path.absolute().as_posix()] + self.config + self.packages
        p = subprocess.Popen(
            cmd,
            cwd=self.root_dir.absolute().as_posix(),
            env=env,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        p.wait()

        assert self.pid
        assert self.log_file

        logger.info(
            f"Connect server started, pid: {self.pid}, logs {self.log_file.absolute().as_posix()}"  # type: ignore
        )

    def _wait_exit(self):
        signal.signal(signal.SIGINT, self.stop)
        signal.signal(signal.SIGTERM, self.stop)
        signal.pause()
        self.daemon_pid_file.unlink()

    def _daemon_and_tail(self):
        daemon_pid = self.pid
        logger.debug(f"Daemon pid {daemon_pid}")
        if not daemon_pid:
            raise DaemonError("No pid to daemon")
        self._tailer = Tailer(self.log_file.as_posix(), daemon_pid).start()  # type: ignore

        def _(daemon_process, interval=1):
            while True:
                time.sleep(interval)
                if not self.pid or self.is_stopping:
                    # No pid file or stopping flag, assume server stopped by self.stop()
                    logger.info("Exit daemon as server has been stopped")
                    self._tailer.stop()
                    return

                if daemon_process.pid != self.pid:
                    # Pid changed, others take over
                    logger.warning("pid has changed, exit")
                    self._tailer.stop()
                    return

                if not daemon_process.is_running():
                    logger.warning("Daemon process is not running, try to restart it")
                    # We have pid file and pid in file is the same as daemon_pid, but no process, restart
                    self._start()
                    daemon_process = psutil.Process(self.pid)
                    # Start a new tailer for new pid
                    self._tailer.stop()
                    self._tailer = Tailer(self.log_file.as_posix(), daemon_process.pid).start()

        self._daemon = threading.Thread(target=_, args=(psutil.Process(daemon_pid),))
        self._daemon.start()

    def _rotate_log(self):
        now = datetime.now()
        count = 1
        for log_file in self.log_dir.glob(self.log_glob_format):
            logger.debug(f"Rotate {log_file.as_posix()}")
            log_file.rename(log_file.with_suffix(f".{now}.log.rotate.{count}"))
            count += 1

    def _launch_daemon(self):
        logger.info("Start daemon connect server")
        if not self.pid:
            raise UnconfigurableError("Connect server is not running")
        self._daemon_and_tail()

    def _stop_connect_server(self, pid: int):
        logger.info("Stop connect server")
        try:
            p = psutil.Process(pid)
        except psutil.NoSuchProcess:
            # Already stopped
            pass
        else:
            # We have the process
            try:
                p.terminate()
                logger.info("Wating for connect server to stop")
                p.wait(30)
            except psutil.TimeoutExpired:
                logger.warning("Timeout for terminate connect server, kill it")
                p.kill()

    def stop(self, sig=None, frame=None):
        self.is_stopping = True
        try:
            pid = self.pid
            if pid:
                self._stop_connect_server(pid)
            if self.pid_file:
                self.pid_file.unlink()
            self._rotate_log()
            if self._daemon:
                self._daemon.join()
        finally:
            self._stopping = False


if __name__ == "__main__":
    d = Daemon()

    d.start_and_daemon()



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License

import os
from datetime import date, datetime
from typing import Dict

import pytest
from pyspark.sql import Row

from sparglim.config.builder import ConfigBuilder
from sparglim.exceptions import UnconfigurableError


@pytest.fixture
def config_builder():
    c = ConfigBuilder()
    yield c
    c.clear()


def verify_spark(spark, config):
    df = spark.createDataFrame(
        [
            Row(a=1, b=2.0, c="string1", d=date(2000, 1, 1), e=datetime(2000, 1, 1, 12, 0)),
            Row(a=2, b=3.0, c="string2", d=date(2000, 2, 1), e=datetime(2000, 1, 2, 12, 0)),
            Row(a=4, b=5.0, c="string3", d=date(2000, 3, 1), e=datetime(2000, 1, 3, 12, 0)),
        ]
    )
    df.show()

    for k, v in config.items():
        assert spark.conf.get(k) == v


def assert_contain(left, right):
    for k, v in right.items():
        if v == None:
            assert k not in left
            continue
        assert k in left
        assert left[k] == v


def patch_env(config_builder: ConfigBuilder, monkeypatch, mapper: Dict[str, str]) -> ConfigBuilder:
    for k, v in mapper.items():
        monkeypatch.setenv(k, v)
    return config_builder.clear()


@pytest.mark.parametrize("mode", ["local", "k8s", "connect_client", "connect_server"])
def test_deploy_mode(config_builder: ConfigBuilder, mode: str, k8s_config_path):
    config_builder = config_builder.clear()
    config_mode = getattr(config_builder, f"config_{mode}")
    if mode == "k8s":
        config_mode(k8s_config_path=k8s_config_path)
    else:
        config_mode()
    if mode != "connect_server":
        assert config_builder.master_configured

    if mode in ["local", "k8s"]:
        with pytest.raises(UnconfigurableError) as e:
            config_mode()


def test_merge(config_builder: ConfigBuilder):
    prev = config_builder.get_all().copy()
    to_merge = {"a": "b"}
    config_builder._merge_config(to_merge)
    assert config_builder.get_all() == {**prev, **to_merge}

    to_cover = {"a": "c"}
    config_builder._merge_config(to_cover)
    assert config_builder.get_all() == {**prev, **to_cover}


def test_config(config_builder: ConfigBuilder):
    prev = config_builder.get_all().copy()
    to_merge = {"spark.app.name": "appname"}
    config_builder.config(to_merge)
    assert config_builder.get_all() == {**prev, **to_merge}

    to_cover = {"spark.app.name": "appname2"}
    config_builder.config(to_cover)
    assert config_builder.get_all() == {**prev, **to_cover}


def test_env(config_builder: ConfigBuilder, monkeypatch):
    config_builder = patch_env(config_builder, monkeypatch, {"SPAGLIM_APP_NAME": "testapp"})
    assert config_builder.get_all()["spark.app.name"] == "testapp"


def test_create(config_builder: ConfigBuilder):
    spark = config_builder.get_or_create()
    config_builder.get_or_create() == spark
    verify_spark(spark, config_builder.get_all())

    old_spark = spark

    config_builder = config_builder.clear()
    spark = config_builder.get_or_create()
    assert old_spark != spark
    verify_spark(spark, config_builder.get_all())


s3_env_cases = [
    (
        {
            "S3_ACCESS_KEY": "s3-access-key",
            "S3_SECRET_KEY": "s3-secret-key",
            "S3_ENTRY_POINT": "s3-entry-point",
            "S3_ENTRY_POINT_REGION": "s3-entry-point-region",
            "S3_PATH_STYLE_ACCESS": "true",
            "S3_MAGIC_COMMITTER": "true",
        },
        {
            "spark.hadoop.fs.s3a.access.key": "s3-access-key",
            "spark.hadoop.fs.s3a.secret.key": "s3-secret-key",
            "spark.hadoop.fs.s3a.endpoint": "s3-entry-point",
            "spark.hadoop.fs.s3a.endpoint.region": "s3-entry-point-region",
            "spark.hadoop.fs.s3a.path.style.access": "true",
            "spark.hadoop.fs.s3a.bucket.all.committer.magic.enabled": "true",
        },
    ),
    (
        {
            "AWS_ACCESS_KEY_ID": "aws-access-key",
            "AWS_SECRET_ACCESS_KEY": "aws-secret-key",
            "AWS_DEFAULT_REGION": "aws-entry-point-region",
        },
        {
            "spark.hadoop.fs.s3a.access.key": "aws-access-key",
            "spark.hadoop.fs.s3a.secret.key": "aws-secret-key",
            "spark.hadoop.fs.s3a.endpoint.region": "aws-entry-point-region",
        },
    ),
]


@pytest.mark.parametrize("env_mapper, expect_mapper", s3_env_cases)
def test_s3(
    config_builder: ConfigBuilder,
    monkeypatch,
    env_mapper: Dict[str, str],
    expect_mapper: Dict[str, str],
):
    config_builder = patch_env(config_builder, monkeypatch, env_mapper)
    config_builder.config_s3()
    assert_contain(config_builder.get_all(), expect_mapper)

    convert_mapper = {
        "spark.hadoop.fs.s3a.access.key": "convert-access-key",
        "spark.hadoop.fs.s3a.secret.key": "convert-secret-key",
    }
    config_builder.config_s3(convert_mapper)
    assert_contain(config_builder.get_all(), {**expect_mapper, **convert_mapper})


# TODO:
# Mock env as incluster
k8s_env_mapper = [
    (
        {
            "SPARGLIM_K8S_NAMESPACE": "namespace",
            "SPARGLIM_K8S_IMAGE": "image",
            "SPARK_EXECUTOR_NUMS": "4",
            "SPARGLIM_K8S_EXECUTOR_LABEL_LIST": "a,b",
            "SPARGLIM_K8S_EXECUTOR_ANNOTATION_LIST": "a,b",
            "SPARGLIM_DRIVER_HOST": "headless-sparglim",
            "SPARGLIM_DRIVER_POD_NAME": "some-spark-app",
            "SPARGLIM_DRIVER_BINDADDRESS": "192.168.1.1",
            "SPARGLIM_K8S_EXECUTOR_REQUEST_CORES": "1",
            "SPARGLIM_K8S_EXECUTOR_LIMIT_CORES": "4",
            "SPARGLIM_EXECUTOR_REQUEST_MEMORY": "1024m",
            "SPARGLIM_EXECUTOR_LIMIT_MEMORY": "2048m",
            "SPARGLIM_K8S_GPU_AMOUNT": "1",
        },
        {
            "spark.kubernetes.namespace": "namespace",
            "spark.kubernetes.container.image": "image",
            "spark.executor.instances": "4",
            "spark.kubernetes.executor.label.a": "true",
            "spark.kubernetes.executor.label.b": "true",
            "spark.kubernetes.executor.annotation.a": "true",
            "spark.kubernetes.executor.annotation.b": "true",
            "spark.driver.host": "headless-sparglim",
            "spark.kubernetes.driver.pod.name": "some-spark-app",
            "spark.driver.bindAddress": "192.168.1.1",
            "spark.kubernetes.executor.cores": "1",
            "spark.kubernetes.executor.limit.cores": "4",
            "spark.executor.memory": "1024m",
            "spark.executor.memoryOverhead": "2048m",
            "spark.executor.resource.gpu.vendor": "nvidia.com",
            "spark.executor.resource.gpu.discoveryScript": "/opt/spark/examples/src/main/scripts/getGpusResources.sh",
            "spark.executor.resource.gpu.amount": "1",
        },
    ),
    (
        {
            # Test default value
        },
        {
            "spark.kubernetes.namespace": "sparglim",
            "spark.kubernetes.executor.label.sparglim-executor": "true",
            "spark.kubernetes.executor.annotation.sparglim-executor": "true",
            "spark.driver.bindAddress": "0.0.0.0",
            "spark.executor.memory": "512m",
            "spark.executor.resource.gpu.vendor": None,
            "spark.executor.resource.gpu.discoveryScript": None,
        },
    ),
]


@pytest.mark.parametrize("env_mapper, expect_mapper", k8s_env_mapper)
def test_k8s(
    config_builder: ConfigBuilder,
    monkeypatch,
    k8s_config_path,
    env_mapper: Dict[str, str],
    expect_mapper: Dict[str, str],
):
    config_builder = patch_env(config_builder, monkeypatch, env_mapper)
    config_builder.config_k8s(k8s_config_path=k8s_config_path)
    assert_contain(config_builder.get_all(), expect_mapper)


def test_k8s_no_config(config_builder: ConfigBuilder):
    with pytest.raises(UnconfigurableError) as e:
        config_builder.config_k8s(k8s_config_path="NOT-EXSIT-FILE")


connect_client_mapper = [
    ({"SPARGLIM_REMOTE": "sc://localhost:9999"}, {"spark.remote": "sc://localhost:9999"}),
    (
        {
            # Test default value
        },
        {"spark.remote": "sc://localhost:15002"},
    ),
]


@pytest.mark.parametrize("env_mapper, expect_mapper", connect_client_mapper)
def test_connect_client(
    config_builder: ConfigBuilder,
    monkeypatch,
    env_mapper: Dict[str, str],
    expect_mapper: Dict[str, str],
):
    config_builder = patch_env(config_builder, monkeypatch, env_mapper)
    config_builder.config_connect_client()
    assert_contain(config_builder.get_all(), expect_mapper)


connect_server_mapper = [
    (
        {
            "SPARGLIM_CONNECT_SERVER_PORT": "9999",
            "SPARGLIM_CONNECT_GRPC_ARROW_MAXBS": "12345",
            "SPARGLIM_CONNECT_GRPC_MAXIM": "67890",
        },
        {
            "spark.connect.grpc.binding.port": "9999",
            "spark.connect.grpc.arrow.maxBatchSize": "12345",
            "spark.connect.grpc.maxInboundMessageSize": "67890",
        },
    ),
    (
        {
            # Test default value
        },
        {},
    ),
]


@pytest.mark.parametrize("env_mapper, expect_mapper", connect_server_mapper)
def test_connect_server(
    config_builder: ConfigBuilder,
    monkeypatch,
    env_mapper: Dict[str, str],
    expect_mapper: Dict[str, str],
):
    config_builder = patch_env(config_builder, monkeypatch, env_mapper)
    config_builder.config_connect_server()
    assert_contain(config_builder.get_all(), expect_mapper)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause
import os
from typing import List, Optional

from IPython.core.magic import Magics, cell_magic, line_magic, magics_class
from pyspark.sql import SparkSession

from sparglim.config.builder import ConfigBuilder
from sparglim.exceptions import UnconfigurableError
from sparglim.log import logger


@magics_class
class SparkMagic(Magics):
    """
    # Enable magic

        ```ipython
        %load_ext sparglim.sql
        ```

    # Example:

        Init a table

        ```python
        from datetime import datetime, date
        from pyspark.sql import Row

        df = spark.createDataFrame([
            Row(a=1, b=2., c='string1', d=date(2000, 1, 1), e=datetime(2000, 1, 1, 12, 0)),
            Row(a=2, b=3., c='string2', d=date(2000, 2, 1), e=datetime(2000, 1, 2, 12, 0)),
            Row(a=4, b=5., c='string3', d=date(2000, 3, 1), e=datetime(2000, 1, 3, 12, 0))
        ])
        df.createOrReplaceTempView("tb")
        ```

        Query it

        ```ipython
        %sql SELECT * FROM tb
        ```

        or
        ```ipython
        %%sql SELECT
        *
        FROM
        tb
        ```
    """

    ENV_MASTER_MODE = "SPARGLIM_SQL_MODE"
    DEFAULT_MODE = "local"

    # TODO: visualize spark dataframe
    def __init__(self, shell=None, **kwargs):
        if "k8s_config_path" in kwargs:
            k8s_config_path = kwargs.pop("k8s_config_path")
        else:
            k8s_config_path = None
        super().__init__(shell=shell, **kwargs)
        self.builder = ConfigBuilder()
        if self.builder.spark:
            logger.info("SparkSession already exists, skip init")
            self.builder.set_runtime_conf("spark.sql.repl.eagerEval.enabled", "true")
            return

        self.builder.config(
            {
                "spark.sql.repl.eagerEval.enabled": "true",
            }
        )

        self.mode: str = os.getenv(self.ENV_MASTER_MODE, self.DEFAULT_MODE)

        if self.mode == "local":
            self.builder.config_local()
        elif self.mode == "connect_client":
            self.builder.config_connect_client()
        elif self.mode == "k8s":
            self.builder.config_k8s(k8s_config_path=k8s_config_path)
        else:
            raise UnconfigurableError(f"Unconfigurable mode {self.mode}")

    @property
    def spark(self) -> SparkSession:
        return self.builder.get_or_create()

    def format_sql(self, sql_lines: List[Optional[str]]):
        # format sql to one line
        lines = ""
        for line in sql_lines:
            line = line or ""
            lines += line.strip() + " "

        return lines.strip()

    def _execute_one(self, sql=str):
        if not sql:
            return
        logger.debug(f"Execute Spark SQL: {sql}")
        return self.spark.sql(sql)

    def _execute(self, sql=str):
        r = None
        for s in sql.split(";"):
            r = self._execute_one(s.strip())
        return r

    @line_magic
    @cell_magic
    def sql(self, line: Optional[str] = None, cell: Optional[str] = None):
        return self._execute(self.format_sql([line, cell]))

    @classmethod
    def clear(self):
        # ConfigBuilder is Singleton
        ConfigBuilder().clear()



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License

import pytest


@pytest.fixture
def cluster_ca_file(tmpdir) -> str:
    cluster_ca_file = tmpdir.join("cluster_ca_file")
    cluster_ca_file.write(
        """
-----BEGIN CERTIFICATE-----
NotARealCertificate==
-----END CERTIFICATE-----
        """
    )
    yield str(cluster_ca_file)
    cluster_ca_file.remove()


@pytest.fixture
def client_ca_file(tmpdir) -> str:
    client_ca_file = tmpdir.join("client_ca_file")
    client_ca_file.write(
        """
-----BEGIN CERTIFICATE-----
NotARealCertificate==
-----END CERTIFICATE-----
        """
    )
    yield str(client_ca_file)
    client_ca_file.remove()


@pytest.fixture
def client_key_file(tmpdir) -> str:
    client_key_file = tmpdir.join("client_key_file")
    client_key_file.write(
        """
-----BEGIN CERTIFICATE-----
NotARealCertificate==
-----END CERTIFICATE-----
        """
    )
    yield str(client_key_file)
    client_key_file.remove()


@pytest.fixture
def k8s_config_path(tmpdir, client_key_file, client_ca_file, cluster_ca_file):
    k8s_config_path = tmpdir.join("k8s_config")
    k8s_config_path.write(
        f"""
apiVersion: v1
clusters:
- cluster:
    certificate-authority: {cluster_ca_file}
    extensions:
    - extension:
        last-update: Sun, 16 Jul 2023 15:22:10 CST
        provider: minikube.sigs.k8s.io
        version: v1.28.0
      name: cluster_info
    server: https://172.27.211.155:8443
  name: minikube
contexts:
- context:
    cluster: minikube
    user: minikube
  name: minikube
current-context: minikube
kind: Config
users:
- name: minikube
  user:
    client-certificate: {client_ca_file}
    client-key: {client_key_file}
"""
    )
    yield k8s_config_path
    k8s_config_path.remove()



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License

from __future__ import annotations

import os
from collections import UserDict
from functools import wraps
from typing import Any, Dict, Iterable, Literal, Optional, Tuple, Union

from sparglim.exceptions import UnconfigurableError
from sparglim.log import logger
from sparglim.utils import get_host_ip

ConfigEnvMapper = Dict[str, Tuple[Union[Iterable, str], Optional[str]]]


class Config(UserDict):
    def __init__(self, dict=None, /, **kwargs):
        self.master_configured: bool = False
        super().__init__(dict, **kwargs)

    def __setitem__(self, key: Any, item: Any) -> None:
        will_config_master = key in ["spark.master", "spark.remote"]
        if self.master_configured and will_config_master:
            raise UnconfigurableError("Spark master/remote already configured, try clear() first")
        if will_config_master:
            self.master_configured = True

        return super().__setitem__(key, item)

    def pretty_format(self) -> str:
        lines = ["{"]
        for k, v in self.items():
            if ("key" in k.lower() or "secret" in k.lower()) and "file" not in k.lower():
                v = "******"
            lines.append(f'  "{k}":"{v}"')
        lines += ["}"]
        return "\n".join(lines)

    def __repr__(self) -> str:
        return self.pretty_format()


class SparkEnvConfiger:
    _basic = {
        "spark.app.name": ("SPAGLIM_APP_NAME", "Sparglim"),
        "spark.submit.deployMode": ("SPAGLIM_DEPLOY_MODE", "client"),
        "spark.scheduler.mode": ("SPARGLIM_SCHEDULER_MODE", "FAIR"),
        "spark.ui.port": ("SPARGLIM_UI_PORT", None),
    }
    _s3 = {
        "spark.hadoop.fs.s3a.access.key": (["S3_ACCESS_KEY", "AWS_ACCESS_KEY_ID"], None),
        "spark.hadoop.fs.s3a.secret.key": (["S3_SECRET_KEY", "AWS_SECRET_ACCESS_KEY"], None),
        "spark.hadoop.fs.s3a.endpoint": ("S3_ENTRY_POINT", None),
        "spark.hadoop.fs.s3a.endpoint.region": (
            ["S3_ENTRY_POINT_REGION", "AWS_DEFAULT_REGION"],
            None,
        ),
        "spark.hadoop.fs.s3a.path.style.access": ("S3_PATH_STYLE_ACCESS", None),
        "spark.hadoop.fs.s3a.bucket.all.committer.magic.enabled": ("S3_MAGIC_COMMITTER", None),
    }
    _kerberos = {
        "spark.kerberos.keytab": ("SPARGIM_KERBEROS_KEYTAB", None),
        "spark.kerberos.principal": ("SPARGIM_KERBEROS_PRINCIPAL", None),
    }
    _local = {
        "spark.master": ("SPARGLIM_MASTER", "local[*]"),
        "spark.driver.memory": ("SPARGLIM_LOCAL_MEMORY", "512m"),
    }
    _connect_client = {
        "spark.remote": ("SPARGLIM_REMOTE", "sc://localhost:15002"),
    }
    _connect_server = {
        "spark.connect.grpc.binding.port": ("SPARGLIM_CONNECT_SERVER_PORT", None),
        "spark.connect.grpc.arrow.maxBatchSize": ("SPARGLIM_CONNECT_GRPC_ARROW_MAXBS", None),
        "spark.connect.grpc.maxInboundMessageSize": ("SPARGLIM_CONNECT_GRPC_MAXIM", None),
    }
    #  FIXME: Does k8s inject more info into pod? So that we can use them directly
    _k8s = {
        # Authenticate will auto config by k8s config file
        # May convert by k8s config file(if exsits)
        "spark.master": ("SPARGLIM_MASTER", "k8s://https://kubernetes.default.svc"),
        "spark.kubernetes.namespace": ("SPARGLIM_K8S_NAMESPACE", None),
        # set SPARGLIM_K8S_IMAGE for image:tag
        "spark.kubernetes.container.image": (
            "SPARGLIM_K8S_IMAGE",
            "wh1isper/spark-executor:3.4.1",
        ),
        "spark.kubernetes.container.image.pullSecrets": ("SPARGLIM_K8S_IMAGE_PULL_SECRETS", None),
        "spark.kubernetes.container.image.pullPolicy": (
            "SPARGLIM_K8S_IMAGE_PULL_POLICY",
            "IfNotPresent",
        ),
        "spark.executor.instances": ("SPARK_EXECUTOR_NUMS", "3"),
        # Expend list from "a,b" -> [a,b], then map them to {label.a: true, label.b: true}
        "spark.kubernetes.executor.label.list": (
            "SPARGLIM_K8S_EXECUTOR_LABEL_LIST",
            "sparglim-executor",
        ),
        "spark.kubernetes.executor.annotation.list": (
            "SPARGLIM_K8S_EXECUTOR_ANNOTATION_LIST",
            "sparglim-executor",
        ),
        # INCLUSTER, work with k8s filedRef.fieldPath, see example
        # TODO: There is a simple example, but no docs yet...
        "spark.driver.host": ("SPARGLIM_DRIVER_HOST", None),
        "spark.driver.bindAddress": ("SPARGLIM_DRIVER_BINDADDRESS", "0.0.0.0"),
        "spark.kubernetes.driver.pod.name": ("SPARGLIM_DRIVER_POD_NAME", None),
        # Config for executor
        "spark.kubernetes.executor.cores": ("SPARGLIM_K8S_EXECUTOR_REQUEST_CORES", None),
        "spark.kubernetes.executor.limit.cores": ("SPARGLIM_K8S_EXECUTOR_LIMIT_CORES", None),
        "spark.executor.memory": ("SPARGLIM_EXECUTOR_REQUEST_MEMORY", "512m"),
        "spark.executor.memoryOverhead": ("SPARGLIM_EXECUTOR_LIMIT_MEMORY", None),
        # GPU
        "spark.executor.resource.gpu.vendor": ("SPARGLIM_K8S_GPU_VENDOR", "nvidia.com"),
        "spark.executor.resource.gpu.discoveryScript": (
            "SPARGLIM_K8S_GPU_DISCOVERY_SCRIPT",
            "/opt/spark/examples/src/main/scripts/getGpusResources.sh",
        ),
        "spark.executor.resource.gpu.amount": ("SPARGLIM_K8S_GPU_AMOUNT", None),
        "spark.rapids.sql.enabled": ("SPARGLIM_RAPIDS_SQL_ENABLED", None)
        # TODO: Mount config
        #       Given that it is highly customizable
        #       Not sure when it will be supported
        #       Welcome PR or discussion here!
    }

    def __init__(self) -> None:
        self.default_config_mapper: ConfigEnvMapper = {
            **self._basic,
            **self._s3,
            **self._kerberos,
        }

        self._config: Config
        self.initialize()

    def initialize(self) -> None:
        self._config: Config = Config(self._config_from_env(self.default_config_mapper))
        logger.debug(f"Initialized config: {self._config}")

    @property
    def master_configured(self) -> bool:
        return self._config.master_configured

    def get_all(self) -> Dict[str, str]:
        return {k: v for k, v in self._config.items() if v is not None}

    def _config_from_env(self, mapper: ConfigEnvMapper) -> Dict[str, Any]:
        config = dict()
        for k, (envs, default) in mapper.items():
            if isinstance(envs, str):
                envs = [envs]
            for env in envs:
                v = os.getenv(env)
                if v:
                    break
            v = v or default

            if v:
                config[k] = v

        return config

    def clear(self) -> SparkEnvConfiger:
        logger.info("Reinitialize config and stop SparkSession")
        if self._spark:
            self._spark.stop()
            self._spark = None
        self.initialize()
        return self

    def _merge_config(self, c: Dict[str, Any]) -> None:
        logger.debug(f"Merge config: {c}")
        self._config.update(**c)
        logger.debug(f"Current config: {self._config}")

    def config(self, c: Dict[str, Any]) -> SparkEnvConfiger:
        self._merge_config(c)
        return self

    def config_s3(self, custom_config: Optional[Dict[str, Any]] = None) -> SparkEnvConfiger:
        if not custom_config:
            custom_config = dict()
        self.config(
            {
                **self._config_from_env(self._s3),
                **custom_config,
            }
        )
        return self

    def config_local(self, custom_config: Optional[Dict[str, Any]] = None) -> SparkEnvConfiger:
        logger.info(f"Config master: local mode")
        if not custom_config:
            custom_config = dict()
        self.config(
            {
                **self._config_from_env(self._local),
                **custom_config,
            }
        )
        return self

    def config_k8s(
        self,
        custom_config: Optional[Dict[str, Any]] = None,
        *,
        k8s_config_path: Optional[str] = None,
    ) -> SparkEnvConfiger:
        try:
            from kubernetes.config.incluster_config import SERVICE_TOKEN_FILENAME

            from sparglim.config.k8s import INCLUSTER, get_k8s_config, get_namespace
        except ImportError:
            raise UnconfigurableError(
                "kubernetes is not installed, try install package `sparglim[k8s]`"
            )

        logger.info(f"Config master: k8s mode")
        if not custom_config:
            custom_config = dict()
        env_config = self._config_from_env(self._k8s)
        if not env_config.get("spark.kubernetes.namespace"):
            env_config["spark.kubernetes.namespace"] = get_namespace()
        if INCLUSTER:
            env_config.setdefault(
                "spark.kubernetes.authenticate.oauthTokenFile", SERVICE_TOKEN_FILENAME
            )
        else:
            env_config.setdefault("spark.driver.host", get_host_ip())

        to_remove_keys = []
        extracted_config = dict()
        for k, v in env_config.items():
            if not k.endswith(".list"):
                continue
            prefix = ".".join(k.split(".")[:-1])
            for item in v.split(","):
                extracted_config[f"{prefix}.{item}"] = "true"
            to_remove_keys.append(k)
        for k in to_remove_keys:
            env_config.pop(k)
        env_config.update(**extracted_config)

        if not env_config.get("spark.executor.resource.gpu.amount"):
            env_config.pop("spark.executor.resource.gpu.vendor")
            env_config.pop("spark.executor.resource.gpu.discoveryScript")

        try:
            url, _, ca, key_file, cert_file = get_k8s_config(k8s_config_path)
        except Exception as e:
            logger.exception(e)
            raise UnconfigurableError("Fail to load k8s config")

        k8s_config = {
            "spark.master": f"k8s://{url}",
            "spark.kubernetes.authenticate.caCertFile": ca,
            "spark.kubernetes.authenticate.clientKeyFile": key_file,
            "spark.kubernetes.authenticate.clientCertFile": cert_file,
        }

        self.config(
            {
                **env_config,
                **k8s_config,
                **custom_config,
            }
        )
        return self

    def config_connect_client(
        self, custom_config: Optional[Dict[str, Any]] = None
    ) -> SparkEnvConfiger:
        if not custom_config:
            custom_config = dict()
        self.config(
            {
                **self._config_from_env(self._connect_client),
                **custom_config,
            }
        )
        return self

    def config_connect_server(
        self,
        mode: Optional[str] = None,
        custom_config: Optional[Dict[str, Any]] = None,
        *,
        k8s_config_path: Optional[str] = None,
    ) -> SparkEnvConfiger:
        if not custom_config:
            custom_config = dict()

        if mode:
            logger.info(f"Config master to connect-server via {mode} mode")
            if mode == "local":
                self.config_local(custom_config)
            elif mode == "k8s":
                self.config_k8s(custom_config, k8s_config_path=k8s_config_path)
            else:
                raise UnconfigurableError(f"Unknown mode: {mode}")
        if k8s_config_path and mode != "k8s":
            logger.warning(f"k8s_config_path has no effort for mode: {mode}")
        logger.info(f"Config connect server")
        self.config(self._config_from_env(self._connect_server))
        return self



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License


import os

import pytest
from IPython.testing.globalipapp import get_ipython

from sparglim.exceptions import UnconfigurableError
from sparglim.sql.magic import SparkMagic

_HERE = os.path.dirname(__file__)
example_file = os.path.join(_HERE, "example/people.json")


ipy = get_ipython()
ipy.run_line_magic("load_ext", "sparglim.sql")


def test_sql():
    ipy.run_line_magic("sql", "SHOW DATABASES")


def test_line():
    ipy.run_line_magic("sql", f'CREATE TABLE tb_people USING json OPTIONS (path "{example_file}")')


def test_cell():
    ipy.run_cell_magic(
        "sql", "CREATE TABLE tb_people2", f' USING json OPTIONS (path "{example_file}")'
    )


def test_mix():
    ipy.run_cell_magic(
        "sql",
        "CREATE TABLE tb_people3",
        f'USING json OPTIONS (path "{example_file}");' "SELECT * FROM tb_people3",
    )
    ipy.run_line_magic("sql", "SHOW TABLES")


@pytest.mark.parametrize("mode", ["local", "k8s", "connect_client", "connect_server"])
def test_env_config(monkeypatch, mode, k8s_config_path):
    monkeypatch.setenv(SparkMagic.ENV_MASTER_MODE, mode)
    SparkMagic.clear()
    if mode == "connect_server":
        with pytest.raises(UnconfigurableError):
            SparkMagic(k8s_config_path=k8s_config_path)
    else:
        SparkMagic(k8s_config_path=k8s_config_path)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from typing import Optional

from pyspark import SparkConf
from pyspark.sql import SparkSession

from sparglim.config.configer import SparkEnvConfiger
from sparglim.exceptions import UnconfigurableError
from sparglim.log import logger
from sparglim.utils import Singleton


class ConfigBuilder(SparkEnvConfiger, metaclass=Singleton):
    def __init__(self):
        super().__init__()
        self.default_master_config = self.config_local
        self._spark: Optional[SparkSession] = None

    @property
    def spark(self) -> Optional[SparkSession]:
        return self._spark

    @property
    def spark_config(self) -> SparkConf:
        config = []
        for k, v in self.get_all().items():
            config.append((k, v))

        return SparkConf().setAll(config)

    def _create(self) -> SparkSession:
        logger.info("Create SparkSession")
        return SparkSession.builder.config(conf=self.spark_config).getOrCreate()

    def get_or_create(self) -> SparkSession:
        if self._spark:
            return self._spark

        if not self.master_configured:
            self.default_master_config()

        self._spark = self._create()
        return self._spark

    def set_runtime_conf(self, k, v):
        if not self._spark:
            raise UnconfigurableError("SparkSession not created yet")
        logger.info(f"Set runtime config: {k}={v}")
        self._spark.conf.set(k, v)



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License

import os
import random
import signal
import string
import sys
import threading
import time
from pathlib import Path

log_dir = Path(os.environ["SPARK_LOG_DIR"])
log_dir.mkdir(parents=True, exist_ok=True)
spark_ident_string = os.environ["SPARK_IDENT_STRING"]
log_file = log_dir / f"spark-{spark_ident_string}-mock-server.out"


def _append_text(self, target: str) -> None:
    with self.open("a") as f:
        f.write(target)


Path.append_text = _append_text


def random_str():
    return "".join(random.choices(string.ascii_letters, k=10)) + "\n"


def start_write_random_str():
    while True:
        log_file.append_text(random_str())
        time.sleep(1)


def stop(sig, frame):
    log_file.append_text("Mock Spark Server Stopped\n")
    sys.exit(0)


def main():
    signal.signal(signal.SIGINT, stop)
    signal.signal(signal.SIGTERM, stop)

    threading.Thread(target=start_write_random_str, daemon=True).start()
    signal.pause()


if __name__ == "__main__":
    main()



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License

from .magic import SparkMagic


def load_ipython_extension(ipython):
    ipython.register_magics(SparkMagic)


def unload_ipython_extension(ipython):
    SparkMagic.clear()



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License
import os
import socket
from pathlib import Path
from typing import Dict, Optional


def port_is_used(port: int) -> bool:
    port = int(port)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.bind(("0.0.0.0", port))
        return False
    except OSError:
        return True


def get_spark_version() -> Optional[str]:
    spark_home = os.getenv("SPARK_HOME")
    if not spark_home:
        return get_pyspark_version()
    jars_dir: Path = Path(spark_home) / "jars"
    for p in jars_dir.glob("spark-core*"):
        # spark-core_2.12-3.4.1.jar -> 3.4.1
        return p.stem.split("_")[-1].split("-")[-1]
    return None


def get_pyspark_version() -> Optional[str]:
    pyspark_version = None
    try:
        import pyspark

        pyspark_version = pyspark.__version__
    except ImportError:
        pass
    return pyspark_version


def get_scala_version() -> Optional[str]:
    spark_home = os.getenv("SPARK_HOME")
    if not spark_home:
        return None
    jars_dir: Path = Path(spark_home) / "jars"
    for p in jars_dir.glob("spark-core*"):
        # spark-core_2.12-3.4.1.jar -> 2.12
        return p.stem.split("_")[-1].split("-")[0]
    return None


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def get_host_ip():
    # Get ip via network connection
    s = None
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        ip = s.getsockname()[0]
    finally:
        if s:
            s.close()
    return ip


def set_if_not_none(config_dict: Dict, k, v):
    if v is not None:
        config_dict[k] = v



from __future__ import annotations

import subprocess
import sys
from typing import Optional

from sparglim.log import logger


class Tailer:
    def __init__(self, file_path: str, pid: int):
        self.cmd = ["tail", "-f", "--pid", f"{pid}", file_path]
        self._task: Optional[subprocess.Popen] = None

    def start(self) -> Tailer:
        if self._task and self._task.poll() is None:
            logger.info("Tailer is already running")
            return self
        logger.info(f"Start tailer by cmd: {self.cmd}")
        self._task = subprocess.Popen(
            self.cmd,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        return self

    def stop(self):
        logger.debug("Try to stop tailer.")
        self._task.kill()
        self._task.wait()
        logger.debug("Tailer stopped")



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License

import os
import subprocess
from pathlib import Path

_HERE = os.path.dirname(__file__)
mock_spark_server = os.path.join(_HERE, "mock_spark_server.py")


def start_mock_server():
    p = subprocess.Popen(
        ["python", mock_spark_server],
        env=os.environ.copy(),
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    pid_dir = Path(os.environ["SPARK_PID_DIR"])
    spark_ident_string = os.environ["SPARK_IDENT_STRING"]
    pid_dir.mkdir(parents=True, exist_ok=True)
    pid_file = pid_dir / f"spark-{spark_ident_string}-mock-server.pid"
    pid_file.write_text(str(p.pid))

    log_dir = Path(os.environ["SPARK_LOG_DIR"])
    log_dir.mkdir(parents=True, exist_ok=True)
    spark_ident_string = os.environ["SPARK_IDENT_STRING"]
    log_file = log_dir / f"spark-{spark_ident_string}-mock-server.out"
    log_file.touch(exist_ok=True)


if __name__ == "__main__":
    print("Starting Mock Spark Server...")
    start_mock_server()



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License
import os
import signal

import click

from sparglim.log import logger
from sparglim.server.daemon import Daemon


@click.command()
@click.option("--mode", default=None)
@click.option("--root_dir", default="./")
def start(mode, root_dir):
    if mode:
        Daemon(mode=mode, root_dir=root_dir).start_and_daemon()
    else:
        Daemon(root_dir=root_dir).start_and_daemon()


@click.command()
@click.option("--root_dir", default="./")
def stop(root_dir="./"):
    logger.disable("sparglim")
    print(f"Stop daemon as root_dir:{root_dir}")
    pid_file = Daemon(root_dir=root_dir).daemon_pid_file
    if pid_file.exists():
        with open(pid_file) as f:
            pid = int(f.read())
        # Stop the daemon by pid
        try:
            print("Kill daemon by pid, send SIGTERM")
            os.kill(pid, signal.SIGTERM)
        except ProcessLookupError:
            pass
    else:
        print(f"No daemon pid file found at {pid_file.as_posix()}, daemon is not running")


@click.group()
def cli():
    pass


cli.add_command(start)
cli.add_command(stop)

if __name__ == "__main__":
    cli()



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License

"""Top-level package for sparglim."""

__author__ = "wh1isper"
__email__ = "9573586@qq.com"
__version__ = "0.1.0rc5"

import findspark  # noqa

try:  # noqa
    findspark.init()  # noqa
except Exception:  # noqa
    pass  # noqa



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License

import os

from kubernetes import config
from kubernetes.client import Configuration

INCLUSTER = os.path.exists("/var/run/secrets/kubernetes.io/serviceaccount/namespace")

if INCLUSTER:
    with open("/var/run/secrets/kubernetes.io/serviceaccount/namespace") as f:
        default_namespace = f.read()
else:
    default_namespace = "spark-app"

NAMESPACE = os.getenv("SPARK_EXECUTOR_NS", default_namespace)


def get_k8s_config(config_file_path=None):
    if INCLUSTER:
        config.load_incluster_config()
    else:
        config_file_path = config_file_path or os.path.expanduser("~/.kube/config")
        config.load_kube_config(config_file_path)
    k8s_config = Configuration.get_default_copy()
    token = k8s_config.api_key.get("authorization")
    host = k8s_config.host
    ca = k8s_config.ssl_ca_cert
    key_file = k8s_config.key_file
    cert_file = k8s_config.cert_file

    return host, token, ca, key_file, cert_file


def get_namespace():
    if not INCLUSTER:
        return "sparglim"
    with open("/var/run/secrets/kubernetes.io/serviceaccount/namespace") as f:
        default_namespace = f.read()
    return default_namespace



#  Copyright (c) 2023 Wh1isper
#  Licensed under the BSD 3-Clause License

import os
import shutil
import signal
import threading
import time
from pathlib import Path

import pytest

from sparglim.server.daemon import Daemon

_HERE = os.path.dirname(__file__)
mock_dir = Path(_HERE) / "mock"


@pytest.fixture
def spark_home():
    yield mock_dir.absolute().as_posix()


@pytest.fixture
def tmppath(tmpdir):
    tmppath = Path(tmpdir)
    yield tmppath
    pid_dir = tmppath / "pids"
    pid_file = pid_dir / "spark-sparglim-connect-server-mock-server.pid"
    if pid_file.exists():
        pid = int(pid_file.read_text())
        try:
            os.kill(pid, 9)
        except Exception:
            pass
    shutil.rmtree(tmppath.as_posix())


@pytest.mark.timeout(35)
def test_daemon(monkeypatch, spark_home, tmppath):
    monkeypatch.setenv("SPARK_HOME", spark_home)

    def _interrupt():
        time.sleep(2)
        os.kill(os.getpid(), signal.SIGTERM)

    threading.Thread(target=_interrupt, daemon=True).start()
    daemon = Daemon(root_dir=tmppath)
    daemon.start_and_daemon()


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])
